;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Test program to determine if reloading CR3 with the same value
; causes a TLB flush.
; Chris Giese	<geezer@execpc.com>	http://my.execpc.com/~geezer
;
; I, the copyright holder of this work, hereby release it into the
; public domain. This applies worldwide. If this is not legally possible:
; I grant any entity the right to use this work for any purpose,
; without any conditions, unless such conditions are required by law.
;
; Updated March 2, 2014:
; - Accessing memory at 1MB without enabling A20 causes a reboot (why?),
;   so I added code to enable (and disable; when done) A20.
;
; Updated ?
; - Touching all 1024 pages of the bottom 4 MB of memory after changing CR3
; - Measuring only the amount of time needed to touch pages,
;   not time needed to load CR3
; - Using timer chip for timing instead of RDTSC,
;   so code should work on 386 and 486 systems
; - Can now run code from DOS or from a bootsector
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

	ORG 0x100		; DOS .COM file

	cld			; string operations go up

; check for DOS PSP
	cmp word [0],0x20CD	; 0xCD 0x20 = "INT 20h"
	jne no_dos

	inc byte [_dos]
	mov ax,0x4300		; if DOS, check for XMS
	int 0x2F
	cmp al,0x80
	jne no_xms

	inc byte [_xms]
	mov ax,0x4310		; get XMS driver address
	push es
		int 0x2F
		mov [_xms_entry + 2],es
		mov [_xms_entry + 0],bx
	pop es
no_xms:
no_dos:
; clear the screen by setting 80x25 text mode
	mov ax,0x0003
	int 0x10

; the 32-bit code writes directly to the screen, so move the BIOS cursor
; down the screen halfway or so
	mov si,_blank_msg
	call puts16

; check for 32-bit CPU
	pushf
		pushf
		pop bx		; old FLAGS -> BX
		mov ax,bx
		xor ah,0x70	; try changing b14 (NT)...
		push ax		; ... or b13:b12 (IOPL)
		popf
		pushf
		pop ax		; new FLAGS -> AX
	popf
	xor ah,bh		; 32-bit CPU if we changed EFLAGS.NT...
	and ah,0x70		; ...or EFLAGS.IOPL
	mov si,_cpu_msg
	je mexit

; check for virtual 8086 mode
	smsw ax			; check MSW.PE = b0 of MSW
	test al,1       	; (MSW = CR0)
	je no_vm86
	mov si,_v86_msg
mexit:
	call puts16		; display message
exit:
	xor ax,ax		; if loaded from DOS...
	or al,[_dos]
	je reboot

	mov ax,0x4C01		; ...DOS terminate
	int 0x21
reboot:
	mov si,_reboot_msg
	call puts16
get_key:
	sti
	mov ah,0		; ...else await key pressed
	int 0x16
	cmp al,0x1B
	jne get_key

	int 0x19		; re-start the boot process

no_vm86:
; need this to return to real mode:
	mov [rm_return + 3],cs

; see if "fast" A20 control supported
	mov ax,0x2403
	int 0x15
	jc no_fast
	cmp ah,0
	jne no_fast
	and bx,2
	je no_fast
	inc byte [_fast_a20]
no_fast:

; set up protected mode Interrupt Descriptor Table (IDT)
	mov cx,(_idt_end - _idt) / 8
	mov bx,_idt
	mov edx,isr0
build_idt:
	mov eax,edx
	mov [bx + 0],ax
	shr eax,16
	mov [bx + 6],ax
	add bx,byte 8		; each interrupt gate is 8 bytes
	add edx,byte (isr1 - isr0)  ; size of each handler stub
	loop build_idt

; set up Global Descriptor Table (GDT) so segment base addresses
; are the same in both protected mode and real mode
	xor ebx,ebx
	mov bx,cs
	shl ebx,4		; EBX = CS * 16
	mov eax,ebx
	shr eax,16

	mov [_gdt_cs32 + 2],bx
	mov [_gdt_cs32 + 4],al
	mov [_gdt_cs32 + 7],ah

	mov [_gdt_cs16 + 2],bx
	mov [_gdt_cs16 + 4],al
	mov [_gdt_cs16 + 7],ah

	xor ebx,ebx
	mov bx,ds
	shl ebx,4		; EBX = DS * 16
	mov eax,ebx
	shr eax,16

	mov [_gdt_ds32 + 2],bx
	mov [_gdt_ds32 + 4],al
	mov [_gdt_ds32 + 7],ah

	mov [_gdt_ds16 + 2],bx
	mov [_gdt_ds16 + 4],al
	mov [_gdt_ds16 + 7],ah

; "fix up" linear addresses in GDT and IDT "pseudo-descriptors"
	lea eax,[ebx + _gdt]
	mov [_gdt_ptr + 2],eax

	lea eax,[ebx + _idt]
	mov [_idt_ptr + 2],eax

; get some memory for page tables. The linear address
; of this memory must be aligned on a page (4K) boundary.
	lea ebp,[ebx + _heap]	; linear address of "_heap"
	add ebp,0x0FFF		; align to 4K
	and bp,0xF000
	mov edi,ebp		; EBP=linear adr (relative to 0)...
	sub edi,ebx		; ...(E)DI=offset (relative to DS base)

; use first 4K of aligned memory for page table that identity-maps
; the bottom 4 MB of memory
	mov cx,1024
	mov eax,3		; b2=0=ring 0, b1=1=writable, b0=1=present
init_ptab:
	stosd
	add eax,0x1000
	loop init_ptab

; use next 4K for page directory (top-level page table) #1
; (at linear address [ebp+4096])
	mov eax,ebp		; physical linear address of page table
	or al,3			; b2=0=ring 0, b1=1=writable, b0=1=present
	stosd
	mov cx,1023		; zero the other 1023 entries
	xor eax,eax
	rep stosd

; use next 4K for page directory #2
; (at linear address [ebp+8192])
	mov eax,ebp		; physical linear address of page table
	or al,3			; b2=0=ring 0, b1=1=writable, b0=1=present
	stosd
	mov cx,1023		; zero the other 1023 entries
	xor eax,eax
	rep stosd

	sub di,4096*3

; now page table        is at linear address (ebp)     , offset [di]
; and page directory #1 is at linear address (ebp+4096), offset [di+4096]
; and page directory #2 is at linear address (ebp+8192), offset [di+8192]

; unmap the 4K just below 1MB to cause a page fault
; (to test my exception handling code)
; This "works" in the sense that my exception handling code is called,
; registers are displayed on the screen, and the system freezes.
; mov dword [di + 255*4],0

; interrupts off, clear NT bit
	push dword 2
	popfd

; get current A20 state
	mov al,1
	call verify_a20
	jne a20_is_on
	xor al,al
a20_is_on:
	mov [_a20_state],al

; put the stack in a known place
	mov esp,_stack

; Enable A20. This code reboots if A20 is disabled, which is strange.
; I'm not writing to extended memory. Why would reads cause a problem
; if they wrap around from address 1MB to address 0?
	mov si,_cant_enable_a20_msg
	call enable_a20
	je mexit

	lgdt [_gdt_ptr]		; load GDTR
	lidt [_idt_ptr]		; load IDTR; to handle CPU exceptions

	mov eax,1		; set CR0.PE
	mov cr0,eax

	jmp CSEL32:pmode		; load all seg regs, starting with CS

; return here on your way back to real mode:
pmode16:
	xor eax,eax		; clear CR0.PE
	mov cr0,eax
rm_return:
	jmp 0:rmode		; code above stores real-mode CS value here
rmode:
	mov ax,cs
	mov ss,ax
	mov ds,ax
	mov es,ax
	mov fs,ax
	mov gs,ax

	lidt [_rm_idt_ptr]

; if A20 was disabled when we started, disable it again
	xor ax,ax
	or al,[_a20_state]
	jne done
	call disable_a20
	je done
	mov si,_cant_disable_a20_msg
	call puts16
done:
	mov si,_done_msg
	jmp mexit

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; name:			set_a20_xms
; action:		enables/disables A20 using XMS (HIMEM.SYS)
; in:			AH=0 to disable A20, AH != 0 to enable A20
; out:			(nothing)
; modifies:		(nothing)
; minimum CPU:		'286 for A20 gate
; notes:
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

set_a20_xms:
	push si			; DEBUG
	push bx
	push ax
		mov si,_xms_msg	; DEBUG
		call puts16	; DEBUG

		or ah,ah
		jne sax1

; I tried "Global disable/enable A20" (XMS functions 3 and 4)
; bu they didn't work. I'm not sure what the difference is...
		mov ah,6	; "Local disable A20"
		jmp short sax2
sax1:
		mov ah,5	; "Local enable A20"
sax2:
		call far [_xms_entry]
		cmp ax,0	; 0=failure, 1=success
	pop ax
	pop bx
	pop si			; DEBUG
	ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; name:			set_a20_fast
; action:		enables/disables A20 using b1 of I/O port 92h
; in:			AH=0 to disable A20, AH != 0 to enable A20
; out:			(nothing)
; modifies:		(nothing)
; minimum CPU:		'286 for A20 gate
; notes:
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

set_a20_fast:
	push si			; DEBUG
	push ax
		mov si,_fast_msg	; DEBUG
		call puts16	; DEBUG

		in al,92h
		or ah,ah
		jne saf1

; chipset bug: do NOT clear b1 if it's already clear
		test al,02h
		je saf3
		and al,0FDh
		jmp short saf2
saf1:
; chipset bug: do NOT set b1 if it's already set
		test al,02h
		jne saf3
		or al,02h
saf2:
		out 92h,al
saf3:
	pop ax
	pop si			; DEBUG
	ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; name:			kbd
; action:		waits until 8042 keyboard controller ready
;			to accept a command or data byte
; modifies:		AL
; out:			(nothing)
; modifies:		(nothing)
; minimum CPU:		'286 for 8042 chip
; notes:
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

kbd0:
	jmp short $+2	; a delay (probably not effective nor necessary)
	in al,60h	; read and discard data/status from 8042
kbd:
	jmp short $+2	; delay
	in al,64h
	test al,1	; output buffer (data _from_ keyboard) full?
	jnz kbd0	; yes, read and discard
	test al,2	; input buffer (data _to_ keyboard) empty?
	jnz kbd		; no, loop
	ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; name:			set_a20_at
; action:		enables/disables A20 at 8042 output port,
;			using D0h (read output port) and D1h
;			(write output port) command bytes to
;			toggle bit b1 of the output port
; in:			AH=0 to disable A20, AH != 0 to enable A20
; out:			(nothing)
; modifies:		(nothing)
; minimum CPU:		'286 for A20 gate
; notes:
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

set_a20_at:
	push si			; DEBUG
	push ax
	pushf
		mov si,_at_msg	; DEBUG
		call puts16	; DEBUG

; Yay, feedback! Chase told me it works better if I shut off interrupts:
		cli
		call kbd
		mov al,0D0h	; 8042 command byte to read output port
		out 64h,al
saa1:
		in al,64h
		test al,1	; output buffer (data _from_ keyboard) full?
		jz saa1	; no, loop

		in al,60h	; read output port
		or ah,ah
		jne saa2
		and al,0FDh	; AND ~2 to disable
		jmp saa3
saa2:
		or al,2		; OR 2 to enable
saa3:
		mov ah,al

		call kbd
		mov al,0D1h	; 8042 command byte to write output port
		out 64h,al

		call kbd
		mov al,ah	; the value to write
		out 60h,al

		call kbd
	popf
	pop ax
	pop si			; DEBUG
	ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; name:			set_a20_vectra
; action:		enables/disables A20 at 8042 output port,
;			using DFh command byte/DDh command byte
; in:			AH=0 to disable A20, AH != 0 to enable A20
; out:			(nothing)
; modifies:		(nothing)
; minimum CPU:		'286 for A20 gate
; notes:
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

set_a20_vectra:
	push si			; DEBUG
	push ax
	pushf
		mov si,_vectra_msg ; DEBUG
		call puts16	; DEBUG

		cli
		call kbd
		or ah,ah
		jne sav1
		mov al,0DDh	; disable
		jmp short sav2
sav1:
		mov al,0DFh	; enable
sav2:
		out 64h,al

		call kbd
	popf
	pop ax
	pop si			; DEBUG
	ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; name:			set_a20_ps2
; action:		enables/disables A20 using INT 15h AX=2400h/2401h
; in:			AH=0 to disable A20, AH != 0 to enable A20
; out:			(nothing)
; modifies:		(nothing)
; minimum CPU:		'286 for A20 gate
; notes:		This call is not supported by my BIOS, but the BIOS
;			enables interrupts!
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

set_a20_ps2:
	push si			; DEBUG
	push ax
	pushf
		mov si,_ps2_msg	; DEBUG
		call puts16	; DEBUG

		or ah,ah
		jne sap1
		mov ax,2400h	; disable
		jmp short sap2
sap1:
		mov ax,2401h	; enable
sap2:
		int 15h
	popf
	pop ax
	pop si			; DEBUG
	ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; name:			verify_a20
; action:		tests if A20 enabled
; in:			(nothing)
; out (A20 enabled):	ZF=0
; out (A20 NOT enabled):ZF=1
; modifies:		(nothing)
; minimum CPU:		'286 for A20 gate
; notes:
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

verify_a20:
	push ax
	push ds
	push es
		xor ax,ax
		mov ds,ax
		dec ax
		mov es,ax

		mov ax,[es:10h]		; read word at FFFF:0010 (1 meg)
		not ax			; 1's complement
		push word [0]		; save word at 0000:0000 (0)
			mov [0],ax	; word at 0 = ~(word at 1 meg)
			mov ax,[0]	; read it back
			cmp ax,[es:10h]	; fail if word at 0 == word at 1 meg
		pop word [0]
	pop es
	pop ds
	pop ax
	ret		; if ZF=1, the A20 gate is NOT enabled

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; name:			enable_a20
; action:		tries to enable A20 using various methods
; in:			[_xms], [_xms_entry], [_fast_a20], [_a20_state]
;			must all be set
; out (A20 enabled):	ZF=0
; out (A20 NOT enabled):ZF=1
; modifies:		(nothing)
; minimum CPU:		'286 for A20 gate
; notes:
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

enable_a20:
	push si			; DEBUG
	push bx
	push ax
		mov ah,1	; enable gate

; is it already enabled?
		xor al,al
		or al,[_a20_state]
		je ea1
		mov si,_already_msg ; DEBUG
		call puts16	; DEBUG
		or al,1		; DEBUG; clear ZF again
		jmp short ea5
ea1:
; use XMS to enable A20
		or al,[_xms]
		je ea2
		call set_a20_xms
		jmp short ea4
ea2:
		or al,[_fast_a20]
		je ea3

; use 'fast' method (port 92h)
		call set_a20_fast
		call verify_a20
		jne ea5
ea3:
; use 'PS/2' method
		call set_a20_ps2
		call verify_a20
		jne ea5

; use 'AT' method
		call set_a20_at
		call verify_a20
		jne ea5

; use 'Vectra' method. This makes Bochs panic
		call set_a20_vectra
ea4:
		call verify_a20
ea5:
	pop ax
	pop bx
	pop si			; DEBUG
	ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; name:			disable_a20
; action:		tries to disable A20 using various methods
; in:			[_xms], [_xms_entry], [_fast_a20], [_a20_state]
;			must be set
; out (A20 enabled):	ZF=0
; out (A20 NOT enabled):ZF=1
; modifies:		(nothing)
; minimum CPU:		'286 for A20 gate
; notes:
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

disable_a20:
	push bx
	push ax
		mov ah,0	; disable gate

; is it already disabled?
		call verify_a20
		je da4

; use XMS to disable A20
		xor al,al
		or al,[_xms]
		je da1
		call set_a20_xms
		jmp short da3
da1:
; use 'fast' method (port 92h)
		or al,[_fast_a20]
		je da2
		call set_a20_fast
		call verify_a20
		je da4
da2:
; use 'PS/2' method
		call set_a20_ps2
		call verify_a20
		je da4

; use 'AT' method
		call set_a20_at
		call verify_a20
		je da4

; use 'Vectra' method. This makes Bochs panic
		call set_a20_vectra
da3:
		call verify_a20
da4:
	pop ax
	pop bx
	ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; name:			puts16
; action:		displays text on screen
; in:			0-terminated string at DS:SI
; out:			(nothing)
; modifies:		(nothing)
; minimum CPU:		8088
; notes:		real-mode (16-bit) code
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

puts16:
	push ax
	push bx
	push si
		mov ah,0x0E	; INT 0x10: teletype output
		xor bx,bx	; video page 0
		jmp .2
.1:
		int 0x10
.2:
		lodsb
		or al,al
		jne .1
	pop si
	pop bx
	pop ax
	ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; 32-BIT CODE
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

	BITS 32

pmode:
	mov ax,DSEL32
	mov ss,ax
	mov ds,ax
	mov es,ax
	mov fs,ax

	mov ax,LSEL
	mov gs,ax

; display "pmode enabled"
	mov esi,_pmode_msg
	call puts32

; disable page size extensions
;	xor eax,eax
;	mov cr4,eax

; enable paging
	lea eax,[ebp + 4096]		; linear address of page dir #1
	mov cr3,eax

	mov eax,0x80000001		; set CR0.PG and CR0.PE
	mov cr0,eax

; do I need an absolute JMP here?? either of these seems to work the same
	jmp short paging_enabled
;	jmp CSEL32:paging_enabled
paging_enabled:

	mov esi,_paging_msg		; display "paging enabled"
	call puts32

;;;;;;;;;;;;; load CR3 with different value ;;;;;;;;;;;;;

	lea ebx,[ebp + 8192]
	mov cr3,ebx

; measure time needed to fault in pages
	call timer			; read start time
	mov ebx,eax

	mov edi,0			; touch all 1024 pages in bottom 4M
	mov ecx,1024
touch1:
	mov eax,[gs:edi]
	add edi,4096
	loop touch1

	call timer			; get elapsed time...
	sub eax,ebx

	mov esi,_diff_msg		; ...and display it...
	call puts32
	mov ebx,10			; ...in decimal
	call wrnum

;;;;;;;;;;;;; load CR3 with same value ;;;;;;;;;;;;;

	mov eax,cr3
	mov cr3,eax

; measure time needed to fault in pages
	call timer			; read start time
	mov ebx,eax

	mov edi,0			; touch all 1024 pages in bottom 4M
	mov ecx,1024
touch2:
	mov eax,[gs:edi]
	add edi,4096
	loop touch2

	call timer			; get elapsed time...
	sub eax,ebx

	mov esi,_same_msg		; ...and display it
	call puts32
	mov ebx,10			; ...in decimal
	call wrnum

;;;;;;;;;;;;; do nothing with CR3 ;;;;;;;;;;;;;

	;mov eax,cr3

; measure time needed to fault in pages
	call timer			; read start time
	mov ebx,eax

	mov edi,0			; touch all 1024 pages in bottom 4M
	mov ecx,1024
touch3:
	mov eax,[gs:edi]
	add edi,4096
	loop touch3

	call timer			; get elapsed time...
	sub eax,ebx

	mov esi,_nop_msg		; ...and display it
	call puts32
	mov ebx,10			; ...in decimal
	call wrnum

	jmp leave_pmode

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; name:			all_ints
; action:		32-bit pmode exception handler
; in:			(nothing)
; out:			(nothing; does not return)
; modifies:		(nothing)
; minimum CPU:		'386
; notes:		32-bit pmode code
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

all_ints:
			mov ax,DSEL32
			mov ss,ax
			mov ds,ax
			mov es,ax
			mov fs,ax

			mov ax,LSEL
			mov gs,ax

			mov esi,_fault1_msg
			call puts32
			mov ebx,16	; display values in hex
			mov ecx,9	; field width

			pop eax		; exception number
			call jwrnum32

		    pop eax		; error code
		    call jwrnum32

		pop eax			; EIP
		call jwrnum32

	    pop eax			; CS
	    call jwrnum32

	pop eax				; EFLAGS
	call jwrnum32

	mov eax,cr2			; page fault address
	call jwrnum32

	mov esi,_fault2_msg
	call puts32

	mov ah,[_gdt_cs32 + 7]		; CS base
	mov al,[_gdt_cs32 + 4]
	shl eax,16
	mov ax,[_gdt_cs32 + 2]
	call jwrnum32

	mov eax,ebp			; EBP
	call jwrnum32

	mov eax,esp			; ESP
	call jwrnum32

	mov eax,cr3			; CR3
	call jwrnum32

; disable paging
leave_pmode:
	mov eax,1		; clear CR0.PG; leave CR0.PE set
	mov cr0,eax
	jmp short paging_disabled
paging_disabled:

; load CS and SS with selectors for segments that have limit 0xFFFF
	mov ax,DSEL16
	mov ss,ax
	jmp CSEL16:pmode16

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; name:			timer
; action:		reads 8243 timer chip free-running counter
; in:			timer count in EAX
; out:			(nothing)
; modifies:		(nothing)
; minimum CPU:		'386
; notes:		32-bit pmode code
;			8243 counter decrements once every 838.09 nsec
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

timer:
	push ebx
	pushf
		xor ebx,ebx
		cli			; interrupts off

		mov al,0		; latch timer value
		out 0x43,al

		in al,0x40		; get LSB
		mov bl,al

		in al,0x40
		mov bh,al

		not ebx			; make it count UP; not down
		mov eax,ebx
	popf
	pop ebx
	ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; name:			putc32
; action:		displays one character on screen
; in:			character in AL
; out:			(nothing)
; modifies:		(nothing)
; minimum CPU:		'386
; notes:		32-bit pmode code
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

putc32:
	push es
	push ebx
	push eax
		cmp al,13		; handle newline
		je .1
		cmp al,' '		; ignore other ASCII control chars
		jb .2
		push eax
			mov ax,LSEL
			mov es,ax

			xor eax,eax
			mov al,[_csr_y]
			mov ah,80
			mul ah
			add al,[_csr_x]
			adc ah,0
			mov ebx,eax
		pop eax
		shl ebx,1		; EBX = (csr_y * 80 + csr_x) * 2

		mov [es:ebx + 0xB8000],al ; store char in video memory

		inc byte [_csr_x]	; advance cursor
		cmp byte [_csr_x],80
		jb .2
.1:
		mov byte [_csr_x],0	; line wrap
		inc byte [_csr_y]
		cmp byte [_csr_y],25
		jb .2
		mov byte [_csr_y],0	; screen wrap (no scrolling)
.2:
	pop eax
	pop ebx
	pop es
	ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; name:			puts32
; action:		displays text on screen
; in:			0-terminated string at DS:ESI
; out:			(nothing)
; modifies:		(nothing)
; minimum CPU:		'386
; notes:		32-bit pmode code
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

puts32:
	push esi
	push eax
		jmp short .2
.1:
		call putc32
.2:
		lodsb
		or al,al
		jne .1
	pop eax
	pop esi
	ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; name:			ltoa
; action:		converts 32-bit unsigned value to string
; in:			32-bit unsigned value in EAX, radix in EBX,
;			ESI -> buffer
; out:			(nothing)
; modifies:		ESI
; minimum CPU:		'386
; notes:		32-bit pmode code
;			ESI must point to zero byte at END of buffer
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

ltoa:
	push edx
	push eax
.1:
		xor edx,edx	; EDX:EAX is dividend...
		div ebx		; ...EBX is divisor...
				; ...EAX is quotient, EDX is remainder
		add dl,'0'	; convert remainder from binary to ASCII
		cmp dl,'9'
		jbe .2
		add dl,('A'-('9'+1))
.2:
		dec esi		; store ASCII remainders in reverse order
		mov [esi],dl
		or eax,eax	; loop until quotient == 0
		jne .1
	pop eax
	pop edx
	ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; name:			wrnum
; action:		displays 32-bit value
; in:			32-bit unsigned value in EAX, radix in EBX
; out:			(nothing)
; modifies:		(nothing)
; minimum CPU:		'386
; notes:		32-bit pmode code
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

wrnum:
	push esi
		mov esi,_num_buf
		call ltoa
		call puts32
	pop esi
	ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; name:			jputs32
; action:		displays string right-justified
;			in minimum field width
; in:			0-terminated string at ESI, field width in ECX
; out:			(nothing)
; modifies:		(nothing)
; minimum CPU:		'386
; notes:
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

jputs32:
	push edi
	push ecx
	push ebx
	push eax
		mov ebx,ecx

; determine length of string
		mov edi,esi
		xor eax,eax
		mov ecx,-1
		repne scasb
		neg ecx
		sub ecx,byte 2

; if string length >= field width, display it normally
		cmp ecx,ebx
		jnb .2

; display (field width - string length) spaces
		neg ecx
		add ecx,ebx
		mov al,' '
.1:
		call putc32
		loop .1
.2:
		call puts32
	pop eax
	pop ebx
	pop ecx
	pop edi
	ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; name:			jwrnum32
; action:		writes 32-bit value right-justified,
;			in minimum field width
; in:			32-bit unsigned value in EAX, radix in EBX,
;			field width in ECX
; out:			(nothing)
; modifies:		(nothing)
; minimum CPU:		'386
; notes:
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

jwrnum32:
	push esi
		mov esi,_num_buf
		call ltoa
		call jputs32
	pop esi
	ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; *** WARNING: Each of the following interrupt-handler stubs
; must be consecutive and they must all be the same size
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

isr0:			; divide error
	push byte 0	; push fake error code
	push byte 0	; push exception number
	jmp all_ints
isr1:			; debug/single step
	push byte 0
	push byte 1
	jmp all_ints
isr2:			; non-maskable interrupt
	push byte 0
	push byte 2
	jmp all_ints
isr3:			; INT3
	push byte 0
	push byte 3
	jmp all_ints
isr4:			; INTO
	push byte 0
	push byte 4
	jmp all_ints
isr5:			; BOUND
	push byte 0
	push byte 5
	jmp all_ints
isr6:			; invalid opcode
	push byte 0
	push byte 6
	jmp all_ints
isr7:			; coprocessor not available
	push byte 0
	push byte 7
	jmp all_ints
isr8:			; double fault
	nop		; ...pushes its own error code
	nop		; (two NOPs = PUSH BYTE NNN = 2 bytes)
	push byte 8
	jmp all_ints
isr9:			; coproc segment overrun (386/486SX only)
	push byte 0
	push byte 9
	jmp all_ints
isr0A:			; bad TSS
	nop
	nop
	push byte 0x0A
	jmp all_ints
isr0B:			; segment not present
	nop
	nop
	push byte 0x0B
	jmp all_ints
isr0C:			; stack fault
	nop
	nop
	push byte 0x0C
	jmp all_ints
isr0D:			; GPF
	nop
	nop
	push byte 0x0D
	jmp all_ints
isr0E:			; page fault
	nop
	nop
	push byte 0x0E
	jmp all_ints
isr0F:
	push byte 0
	push byte 0x0F
	jmp all_ints
isr10:			; FP exception/coprocessor error
	push byte 0
	push byte 0x10
	jmp all_ints
isr11:			; alignment check (486+ only)
	push byte 0
	push byte 0x11
	jmp all_ints
isr12:			; machine check (Pentium+ only)
	push byte 0
	push byte 0x12
	jmp all_ints
isr13:
	push byte 0
	push byte 0x13
	jmp all_ints
isr14:
	push byte 0
	push byte 0x14
	jmp all_ints
isr15:
	push byte 0
	push byte 0x15
	jmp all_ints
isr16:
	push byte 0
	push byte 0x16
	jmp all_ints
isr17:
	push byte 0
	push byte 0x17
	jmp all_ints
isr18:
	push byte 0
	push byte 0x18
	jmp all_ints
isr19:
	push byte 0
	push byte 0x19
	jmp all_ints
isr1A:
	push byte 0
	push byte 0x1A
	jmp all_ints
isr1B:
	push byte 0
	push byte 0x1B
	jmp all_ints
isr1C:
	push byte 0
	push byte 0x1C
	jmp all_ints
isr1D:
	push byte 0
	push byte 0x1D
	jmp all_ints
isr1E:
	push byte 0
	push byte 0x1E
	jmp all_ints
isr1F:
	push byte 0
	push byte 0x1F
	jmp all_ints


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; DATA
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

_blank_msg:
	db 13, 10,	13, 10,		13, 10,		13, 10
	db 13, 10,	13, 10,		13, 10,		13, 10
	db 13, 10,	13, 10,		13, 10,		13, 10
	db 13, 10,	13, 10, 0
_cpu_msg:
	db "Error: 32-bit CPU required (386SX or better)", 13, 10, 0
_v86_msg:
	db "Error: CPU in Virtual-8086 mode (Windows DOS box or EMM386 loaded?)"
	db 13, 10, 0
_cant_enable_a20_msg:
	db "Error: can't enable A20 gate", 13, 10, 0
_cant_disable_a20_msg:
	db "Error: can't disable A20 gate", 13, 10, 0
_xms_msg:			; DEBUG
	db "Controlling A20 using 'XMS' method...", 13, 10, 0
_fast_msg:			; DEBUG
	db "Controlling A20 using 'fast' method...", 13, 10, 0
_at_msg:				; DEBUG
	db "Controlling A20 using 'AT' method...", 13, 10, 0
_vectra_msg:			; DEBUG
	db "Controlling A20 using 'Vectra' method...", 13, 10, 0
_ps2_msg:			; DEBUG
	db "Controlling A20 using 'PS/2' method...", 13, 10, 0
_already_msg:
	db "A20 gate already enabled", 13, 10, 0

_pmode_msg:
	db "Now in 32-bit protected mode", 13, 10, 0
_paging_msg:
	db "Paging enabled", 13, 10, 0
_diff_msg:
	db "Clock count after loading CR3 with different value: ", 0
_same_msg:
	db 13, 10, "Clock count after loading CR3 with SAME value: ", 0
_nop_msg:
	db 13, 10, "Clock count after doing nothing with CR3: ", 0
_done_msg:
	db "Back to real mode, I hope...", 13, 10, 0
_reboot_msg:
	db "Press Esc to reboot", 13, 10, 0
_fault1_msg:
	db 13, 10
	db "CPU exception occurred                        Page", 13, 10
	db " Except.  Error                               fault", 13, 10
	db " number   code     EIP      CS       EFLAGS   address", 13, 10
	db " -------- -------- -------- -------- -------- --------", 13, 10, 0
_fault2_msg:
	db 13, 10
	db " CS base  EBP      ESP      CR3", 13, 10
	db " -------- -------- -------- --------", 13, 10, 0

_gdt:
	dd 0, 0			; NULL descriptor

LSEL equ ($ - _gdt)		; 32-bit LINEAR DATA segment selector
	dw 0xFFFF		; limit 15:0
	dw 0			; base 15:0
	db 0			; base 23:16
	db 0x92			; present, ring 0, data, expand-up, writable
	db 0xCF			; 32-bit segment, page-granular limit 4GB-1
	db 0			; base 31:24

_gdt_cs32:			; 32-bit CODE segment descriptor...
CSEL32 equ ($ - _gdt)		; ...and selector
	dw 0xFFFF		; limit 15:0
	dw 0			; base 15:0 (set above)
	db 0			; base 23:16 (set above)
	db 0x9A			; present, ring 0, code, non-conforming, readable
	db 0xCF			; page-granular limit 0xFFFFF = 4GB-1
	db 0			; base 31:24 (set above)

_gdt_ds32:			; 32-bit DATA segment descriptor...
DSEL32 equ ($ - _gdt)		; ...and selector
	dw 0xFFFF		; limit 15:0
	dw 0			; base 15:0 (set above)
	db 0			; base 23:16 (set above)
	db 0x92			; present, ring 0, data, expand-up, writable
	db 0xCF			; page-granular limit 0xFFFFF = 4GB-1
	db 0			; base 31:24 (set above)

_gdt_cs16:			; 16-bit CODE segment descriptor...
CSEL16 equ ($ - _gdt)		; ...and selector
	dw 0xFFFF		; limit 15:0
	dw 0			; base 15:0 (set above)
	db 0			; base 23:16 (set above)
	db 0x9A			; present, ring 0, code, non-conforming, readable
	db 0			; 16-bit segment, byte-granular limit 64K-1
	db 0			; base 31:24 (set above)

_gdt_ds16:			; 16-bit DATA segment descriptor...
DSEL16 equ ($ - _gdt)		; ...and selector
	dw 0xFFFF		; limit 15:0
	dw 0			; base 15:0 (set above)
	db 0			; base 23:16 (set above)
	db 0x92			; present, ring 0, data, expand-up, writable
	db 0			; 16-bit segment, byte-granular limit 64K-1
	db 0			; base 31:24 (set above)
_gdt_end:

_gdt_ptr:
	dw _gdt_end - _gdt - 1	; GDT limit
	dd _gdt			; GDT linear adr (fixed up above)

_idt:
	dw 0			; low 16 bits of EIP (handler offset)
	dw CSEL32		; CS (handler segment)
	db 0			; not used
	db 0x8E			; present, ring 0, 32-bit interrupt gate
	dw 0			; high 16 bits of EIP (handler offset)

	dw 0, CSEL32, 0x8E00, 0
	dw 0, CSEL32, 0x8E00, 0
	dw 0, CSEL32, 0x8E00, 0

	dw 0, CSEL32, 0x8E00, 0
	dw 0, CSEL32, 0x8E00, 0
	dw 0, CSEL32, 0x8E00, 0
	dw 0, CSEL32, 0x8E00, 0

	dw 0, CSEL32, 0x8E00, 0
	dw 0, CSEL32, 0x8E00, 0
	dw 0, CSEL32, 0x8E00, 0
	dw 0, CSEL32, 0x8E00, 0

	dw 0, CSEL32, 0x8E00, 0
	dw 0, CSEL32, 0x8E00, 0
	dw 0, CSEL32, 0x8E00, 0
	dw 0, CSEL32, 0x8E00, 0

	dw 0, CSEL32, 0x8E00, 0
	dw 0, CSEL32, 0x8E00, 0
	dw 0, CSEL32, 0x8E00, 0
	dw 0, CSEL32, 0x8E00, 0

	dw 0, CSEL32, 0x8E00, 0
	dw 0, CSEL32, 0x8E00, 0
	dw 0, CSEL32, 0x8E00, 0
	dw 0, CSEL32, 0x8E00, 0

	dw 0, CSEL32, 0x8E00, 0
	dw 0, CSEL32, 0x8E00, 0
	dw 0, CSEL32, 0x8E00, 0
	dw 0, CSEL32, 0x8E00, 0

	dw 0, CSEL32, 0x8E00, 0
	dw 0, CSEL32, 0x8E00, 0
	dw 0, CSEL32, 0x8E00, 0
	dw 0, CSEL32, 0x8E00, 0
_idt_end:

_idt_ptr:
	dw _idt_end - _idt - 1
	dd _idt

_rm_idt_ptr:
	dw 0x3FF
	dd 0

	times 256 db 0
_num_buf:
	db 0
_csr_x:
	db 0
_csr_y:
	db 0
_dos:
	db 0
_xms:
	db 0
_xms_entry:
	dw 0, 0
_fast_a20:
	db 0
_a20_state:
	db 0

	times 512 db 0
_stack:

_heap:
